# oscillode
# Copyright (C) 2024 oscillode Project Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Basic 'single-source-of-truth' static and dynamic information about this addon.

Information is mined both from `bpy`, and from the addon's bundled manifest file.
"""

import tomllib
import typing as typ
from pathlib import Path

import bpy

PATH_ADDON_ROOT = Path(__file__).resolve().parent.parent

with (PATH_ADDON_ROOT / 'blender_manifest.toml').open('rb') as f:
	MANIFEST = tomllib.load(f)

NAME: str = MANIFEST['id']
VERSION: str = MANIFEST['version']

####################
# - Assets
####################
PATH_ASSETS: Path = PATH_ADDON_ROOT / 'assets'
PATH_CACHE: Path = Path(bpy.utils.extension_path_user(__package__, create=True))


####################
# - Dynamic Addon Information
####################
def operator(
	name: str, *operator_args: list[typ.Any], **operator_kwargs: dict[str, typ.Any]
) -> None:
	"""Convenienve method to call an operator from this addon.

	This avoids having to use `bpy.ops.<addon_name>.operator_name`, which isn't generic.

	Only operators from this addon may be called using this method.

	Raises:
		ValueError: If an addon from outside this operator is attempted to be used.
	"""
	# Parse Operator Name
	operator_namespace, operator_name = name.split('.')
	if operator_namespace != NAME:
		msg = f'Tried to call operator {operator_name}, but addon operators may only use the addon operator namespace "{operator_namespace}.<name>"'
		raise ValueError(msg)

	# Run Operator
	operator = getattr(getattr(bpy.ops, NAME), operator_name)
	operator(*operator_args, **operator_kwargs)
	## TODO: Can't we constrain 'name' to be an OperatorType somehow?


def prefs() -> bpy.types.AddonPreferences | None:
	"""Retrieve the preferences of this addon, if they are available yet.

	Notes:
		While registering the addon, one may wish to use the addon preferences.
		This isn't possible - not even for default values.

		Either a bad idea is at work, or `oscillode.utils.init_settings` should be consulted until the preferences are available.

	Returns:
		The addon preferences, if the addon is registered and loaded - otherwise None.
	"""
	addon = bpy.context.preferences.addons.get(NAME)
	if addon is None:
		return None
	return addon.preferences


####################
# - Logging Info
####################
DEFAULT_LOG_PATH = PATH_ADDON_ROOT / 'addon.log'
## By default, the addon file log writes to the addon dir.
## The initial .log_level contents are written when packing the .zip.
## Subsequent changes are managed by nodeps.utils.simple_logger.py.
