# oscillode
# Copyright (C) 2024 oscillode Project Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# blender_maxwell
# Copyright (C) 2024 blender_maxwell Project Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Implements `ModeSolverNode`."""

import typing as typ

import tidy3d as td

from blender_maxwell.utils import bl_cache, logger
from blender_maxwell.utils import sympy_extra as spux
from blender_maxwell.utils.frozendict import frozendict

from ... import contracts as ct
from ... import sockets
from .. import base, events

log = logger.get(__name__)

FK = ct.FlowKind
FS = ct.FlowSignal
MT = spux.MathType
PT = spux.PhysicalType


class ModeSolverNode(base.MaxwellSimNode):
	"""Definition of a complete FDTD simulation, including boundary conditions, domain, sources, structures, monitors, and other configuration."""

	node_type = ct.NodeType.ModeSolver
	bl_label = 'Mode Solver'
	use_sim_node_name = True

	####################
	# - Sockets
	####################
	input_sockets: typ.ClassVar = {
		'Solved Modes': sockets.MaxwellModeSocketDef(),
		'Inj Mode': sockets.ExprSocketDef(
			mathtype=MT.Integer,
			abs_min=0,
			default_value=0,
		),
		'Guess neff': sockets.ExprSocketDef(
			mathtype=MT.Integer,
			abs_min=1,
		),
		## TODO: Spherical, bend, mode tracking, group index
	}
	output_sockets: typ.ClassVar = {
		'Angled Source': sockets.MaxwellSourceSocketDef(active_kind=FK.Value),
	}

	####################
	# - Properties: UI
	####################
	injection_axis: ct.SimSpaceAxis = bl_cache.BLField(ct.SimSpaceAxis.X)
	injection_direction: ct.SimAxisDir = bl_cache.BLField(ct.SimAxisDir.Plus)

	####################
	# - FlowKind.Value
	####################
	@events.computes_output_socket(
		'Sim',
		kind=FK.Value,
		# Loaded
		props={'active_socket_set'},
		outscks_kinds={'Sim': {FK.Func, FK.Params}},
		all_loose_input_sockets=True,
		loose_input_sockets_kind={FK.Func, FK.Params},
	)
	def compute_value(
		self, props, loose_input_sockets, output_sockets
	) -> td.Simulation | FS:
		"""Compute the particular value of the simulation domain from strictly non-symbolic inputs."""
		func = output_sockets['Sim'][FK.Func]
		params = output_sockets['Sim'][FK.Params]

		has_func = not FS.check(func)
		has_params = not FS.check(params)

		active_socket_set = props['active_socket_set']
		if has_func and has_params and active_socket_set == 'Single':
			symbol_values = {
				sym: events.realize_known(loose_input_sockets[sym.name])
				for sym in params.sorted_symbols
			}

			return func.realize(
				params,
				symbol_values=frozendict(
					{
						sym: events.realize_known(loose_input_sockets[sym.name])
						for sym in params.sorted_symbols
					}
				),
			).updated_copy(
				attrs=dict(
					ct.SimMetadata(
						realizations=ct.SimRealizations(
							syms=tuple(symbol_values.keys()),
							vals=tuple(symbol_values.values()),
						)
					)
				)
			)
		return FS.FlowPending


####################
# - Blender Registration
####################
BL_REGISTER = [
	ModeSolverNode,
]
BL_NODES = {ct.NodeType.ModeSolver: (ct.NodeCategory.MAXWELLSIM_SIMS)}
